import torch
import torch.nn.functional as F
import numpy as np
from scipy import interpolate


def load_ckpt(model, path):
    """Load checkpoint"""
    state_dict = torch.load(path, map_location=torch.device("cpu"))
    model.load_state_dict(state_dict, strict=False)


def resize_data(img1, img2, flow, factor=1.0):
    _, _, h, w = img1.shape
    h = int(h * factor)
    w = int(w * factor)
    img1 = F.interpolate(img1, (h, w), mode="area")
    img2 = F.interpolate(img2, (h, w), mode="area")
    flow = F.interpolate(flow, (h, w), mode="area") * factor
    return img1, img2, flow


class InputPadder:
    """Pads images such that dimensions are divisible by 8"""

    def __init__(self, dims, mode="sintel"):
        self.ht, self.wd = dims[-2:]
        pad_ht = (((self.ht // 8) + 1) * 8 - self.ht) % 8
        pad_wd = (((self.wd // 8) + 1) * 8 - self.wd) % 8
        if mode == "sintel":
            self._pad = [
                pad_wd // 2,
                pad_wd - pad_wd // 2,
                pad_ht // 2,
                pad_ht - pad_ht // 2,
            ]
        else:
            self._pad = [pad_wd // 2, pad_wd - pad_wd // 2, 0, pad_ht]

    def pad(self, *inputs):
        return [F.pad(x, self._pad, mode="replicate") for x in inputs]

    def unpad(self, x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]
        return x[..., c[0] : c[1], c[2] : c[3]]


def forward_interpolate(flow):
    dtype = flow.dtype
    device = flow.device

    flow = flow.detach().cpu().numpy()
    dx, dy = flow[0], flow[1]

    ht, wd = dx.shape
    x0, y0 = np.meshgrid(np.arange(wd), np.arange(ht))

    x1 = x0 + dx
    y1 = y0 + dy

    x1 = x1.reshape(-1)
    y1 = y1.reshape(-1)
    dx = dx.reshape(-1)
    dy = dy.reshape(-1)

    valid = (x1 > 0) & (x1 < wd) & (y1 > 0) & (y1 < ht)
    x1 = x1[valid]
    y1 = y1[valid]
    dx = dx[valid]
    dy = dy[valid]

    flow_x = interpolate.griddata(
        (x1, y1), dx, (x0, y0), method="nearest", fill_value=0
    )

    flow_y = interpolate.griddata(
        (x1, y1), dy, (x0, y0), method="nearest", fill_value=0
    )

    flow = np.stack([flow_x, flow_y], axis=0)
    return torch.from_numpy(flow).to(dtype=dtype, device=device)


def bilinear_sampler(img, coords, mode="bilinear", mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.to(dtype=coords.dtype)

    return img


def coords_grid(batch, ht, wd, dtype, device):
    coords = torch.meshgrid(
        torch.arange(ht, dtype=dtype, device=device),
        torch.arange(wd, dtype=dtype, device=device),
        indexing="ij",
    )
    coords = torch.stack(coords[::-1], dim=0)
    return coords[None].repeat(batch, 1, 1, 1)


def upflow8(flow, mode="bilinear"):
    new_size = (8 * flow.shape[2], 8 * flow.shape[3])
    return 8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)


def transform(T, p):
    assert T.shape == (4, 4)
    return np.einsum("H W j, i j -> H W i", p, T[:3, :3]) + T[:3, 3]


def from_homog(x):
    return x[..., :-1] / x[..., [-1]]


def reproject(depth1, pose1, pose2, K1, K2):
    H, W = depth1.shape
    x, y = np.meshgrid(np.arange(W), np.arange(H), indexing="xy")
    img_1_coords = np.stack((x, y, np.ones_like(x)), axis=-1).astype(np.float64)
    cam1_coords = np.einsum(
        "H W, H W j, i j -> H W i", depth1, img_1_coords, np.linalg.inv(K1)
    )
    rel_pose = np.linalg.inv(pose2) @ pose1
    cam2_coords = transform(rel_pose, cam1_coords)
    return from_homog(np.einsum("H W j, i j -> H W i", cam2_coords, K2))


def induced_flow(depth0, depth1, data):
    H, W = depth0.shape
    coords1 = reproject(depth0, data["T0"], data["T1"], data["K0"], data["K1"])
    x, y = np.meshgrid(np.arange(W), np.arange(H), indexing="xy")
    coords0 = np.stack([x, y], axis=-1)
    flow_01 = coords1 - coords0

    H, W = depth1.shape
    coords1 = reproject(depth1, data["T1"], data["T0"], data["K1"], data["K0"])
    x, y = np.meshgrid(np.arange(W), np.arange(H), indexing="xy")
    coords0 = np.stack([x, y], axis=-1)
    flow_10 = coords1 - coords0

    return flow_01, flow_10


def check_cycle_consistency(flow_01, flow_10):
    flow_01 = torch.from_numpy(flow_01).permute(2, 0, 1)[None]
    flow_10 = torch.from_numpy(flow_10).permute(2, 0, 1)[None]
    H, W = flow_01.shape[-2:]
    coords = coords_grid(1, H, W, flow_01.dtype, flow_01.device)
    coords1 = coords + flow_01
    flow_reprojected = bilinear_sampler(flow_10, coords1.permute(0, 2, 3, 1))
    cycle = flow_reprojected + flow_01
    cycle = torch.norm(cycle, dim=1)
    mask = (cycle < 0.1 * min(H, W)).float()
    return mask[0].numpy()



import torch
import torch.nn.functional as F
import torch.nn as nn

from ...utils.utils import coords_grid


def initialize_flow(img):
    """Flow is represented as difference between two means flow = mean1 - mean0"""
    N, C, H, W = img.shape
    mean = coords_grid(N, H, W).to(img.device)
    mean_init = coords_grid(N, H, W).to(img.device)

    # optical flow computed as difference: flow = mean1 - mean0
    return mean, mean_init


def generate_window_grid(h_min, h_max, w_min, w_max, len_h, len_w, dtype, device):
    assert device is not None

    x, y = torch.meshgrid(
        [
            torch.linspace(w_min, w_max, len_w, dtype=dtype, device=device),
            torch.linspace(h_min, h_max, len_h, dtype=dtype, device=device),
        ],
        indexing="ij",
    )
    grid = torch.stack((x, y), -1).transpose(0, 1)  # [H, W, 2]

    return grid


def normalize_coords(coords, h, w):
    # coords: [B, H, W, 2]
    c = torch.Tensor([(w - 1) / 2.0, (h - 1) / 2.0]).to(
        dtype=coords.dtype, device=coords.device
    )
    return (coords - c) / c  # [-1, 1]


class quater_upsampler(nn.Module):
    def __init__(self):
        super(quater_upsampler, self).__init__()

        hidden_dim = 64

        self.flow_encoder = nn.Sequential(
            nn.Conv2d(2, hidden_dim, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(hidden_dim, hidden_dim - 2, 3, padding=1),
        )

        self.corr_encoder = nn.Sequential(
            nn.Conv2d(9, hidden_dim, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(hidden_dim, hidden_dim, 3, padding=1),
        )

        self.mask_head = nn.Sequential(
            nn.Conv2d(hidden_dim * 2 + 128, 128, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 16 * 9, 1, padding=0),
        )

        self.flow_head = nn.Sequential(
            nn.Conv2d(hidden_dim * 2 + 128, 128, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 2, 1, padding=0),
        )

    def get_local_cost(self, coords, feature0, feature1, local_radius=1):
        b, c, h, w = feature0.size()

        coords = coords.view(b, 2, -1).permute(0, 2, 1)  # [B, H*W, 2]

        local_h = 2 * local_radius + 1
        local_w = 2 * local_radius + 1

        window_grid = generate_window_grid(
            -local_radius,
            local_radius,
            -local_radius,
            local_radius,
            local_h,
            local_w,
            device=feature0.device,
        )  # [2R+1, 2R+1, 2]
        window_grid = window_grid.reshape(-1, 2).repeat(
            b, 1, 1, 1
        )  # [B, 1, (2R+1)^2, 2]
        sample_coords = coords.unsqueeze(-2) + window_grid  # [B, H*W, (2R+1)^2, 2]

        # normalize coordinates to [-1, 1]
        sample_coords_norm = normalize_coords(sample_coords, h, w)  # [-1, 1]
        window_feature = F.grid_sample(
            feature1, sample_coords_norm, padding_mode="zeros", align_corners=True
        ).permute(
            0, 2, 1, 3
        )  # [B, H*W, C, (2R+1)^2]
        feature0_view = feature0.permute(0, 2, 3, 1).view(
            b, h * w, 1, c
        )  # [B, H*W, 1, C]

        corr = torch.matmul(feature0_view, window_feature).view(b, h * w, -1) / (
            c**0.5
        )  # [B, H*W, (2R+1)^2]

        corr = corr.view(b, h, w, -1).permute(0, 3, 1, 2)

        return corr

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/4, W/4, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 4, 4, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(4 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 4 * H, 4 * W)

    def forward(self, flow, context_quater, feat_s_quater, feat_t_quater, r=1):
        coords0, _ = initialize_flow(flow)
        coords1 = coords0 + flow
        local_cost = self.get_local_cost(
            coords1, feat_s_quater, feat_t_quater, local_radius=r
        )

        corr_feat = self.corr_encoder(local_cost)
        feat = torch.cat(
            [flow, self.flow_encoder(flow), corr_feat, context_quater], dim=1
        )

        delta_flow = self.flow_head(feat)
        mask = self.mask_head(feat)

        flow = self.upsample_flow(flow + delta_flow, mask)

        return flow



"""Abstract model to be used as a parent of concrete optical flow networks.

In order to use this module, the concrete model should just need to:
1. Specify the network structure in their __init__() method and call super().__init__() with the required args.
2. Implement a forward method which receives inputs and outputs as specified. See the forward method docs for more details.
3. Optionally, define a loss function, if the model is able to be trained.
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from abc import abstractmethod
import math
from typing import Any, Callable, Dict, Optional, Tuple, Union

import lightning.pytorch as pl
from loguru import logger
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from ptlflow.utils.utils import InputPadder, InputScaler
from ptlflow.utils.utils import bgr_val_as_tensor
from ptlflow.utils.flow_metrics import FlowMetrics

DATASET_MAIN_METRIC = {
    "autoflow": "epe",
    "flyingchairs": "epe",
    "flyingchairs2": "epe",
    "flyingthings3d": "epe",
    "flyingthings3dsubset": "epe",
    "hd1k": "flall",
    "kitti_2012": "flall",
    "kitti_2015": "flall",
    "kubric": "epe",
    "middlebury": "epe",
    "middleburyst": "epe",
    "monkaa": "epe",
    "sintel_clean": "epe",
    "sintel_final": "epe",
    "spring": "px1",
    "tartanair_easy": "epe",
    "tartanair_hard": "epe",
    "viper": "wauc",
}


class BaseModel(pl.LightningModule):
    """A base abstract optical flow model."""

    def __init__(
        self,
        output_stride: int,
        loss_fn: Optional[Callable] = None,
        lr: Optional[float] = None,
        wdecay: Optional[float] = None,
        warm_start: bool = False,
    ) -> None:
        """Initialize BaseModel.

        Parameters
        ----------
        args : Namespace
            A namespace with the required arguments. Typically, this can be gotten from add_model_specific_args().
        loss_fn : Callable
            A function to be used to compute the loss for the training. The input of this function must match the output of the
            forward() method. The output of this function must be a tensor with a single value.
        output_stride : int
            How many times the output of the network is smaller than the input.
        """
        super(BaseModel, self).__init__()

        self.output_stride = output_stride
        self.loss_fn = loss_fn
        self.lr = lr
        self.wdecay = wdecay
        self.warm_start = warm_start

        self.train_size = None
        self.train_avg_length = None

        self.extra_params = None

        self.train_metrics = FlowMetrics(prefix="train/")
        self.val_metrics = nn.ModuleList()
        self.val_dataset_names = []

        self.last_inputs = None
        self.last_predictions = None

        self.prev_preds = None

        self.has_trained_on_ptlflow = False

        self.has_logged_main_val_metric_message = False

        self.save_hyperparameters(
            ignore=["loss_fn"],
        )

    @property
    def train_size(self):
        return self._train_size

    @train_size.setter
    def train_size(self, value):
        if value is not None:
            assert isinstance(value, (tuple, list))
            assert len(value) == 2
            assert isinstance(value[0], int) and isinstance(value[1], int)
        self._train_size = value

    def add_extra_param(self, name, value):
        if self.extra_params is None:
            self.extra_params = {}
        self.extra_params[name] = value

    def preprocess_images(
        self,
        images: torch.Tensor,
        bgr_add: Union[float, Tuple[float, float, float], np.ndarray, torch.Tensor] = 0,
        bgr_mult: Union[
            float, Tuple[float, float, float], np.ndarray, torch.Tensor
        ] = 1,
        bgr_to_rgb: bool = False,
        image_resizer: Optional[Union[InputPadder, InputScaler]] = None,
        resize_mode: str = "pad",
        target_size: Optional[Tuple[int, int]] = None,
        pad_mode: str = "replicate",
        pad_value: float = 0.0,
        pad_two_side: bool = True,
        interpolation_mode: str = "bilinear",
        interpolation_align_corners: bool = True,
    ) -> Tuple[torch.Tensor, Union[InputPadder, InputScaler]]:
        """Applies basic pre-processing to the images.

        The pre-processing is done in this order:
        1. images = images + bgr_add
        2. images = images * bgr_mult
        3. (optional) Convert BGR channels to RGB
        4. Pad or resize the input to the closest larger size multiple of self.output_stride

        Parameters
        ----------
        images : torch.Tensor
            A tensor with at least 3 dimensions in this order: [..., 3, H, W].
        bgr_add : Union[float, Tuple[float, float, float], np.ndarray, torch.Tensor], default 0
            BGR values to be added to the images. It can be a single value, a triple, or a tensor with a shape compatible with images.
        bgr_mult : Union[float, Tuple[float, float, float], np.ndarray, torch.Tensor], default 1
            BGR values to be multiplied by the images. It can be a single value, a triple, or a tensor with a shape compatible with images.
        bgr_to_rgb : bool, default False
            If True, flip the channels to convert from BGR to RGB.
        image_resizer : Optional[Union[InputPadder, InputScaler]]
            An instance of InputPadder or InputScaler that will be used to resize the images.
            If not provided, a new one will be created based on the given resize_mode.
        resize_mode : str, default "pad"
            How to resize the input. Accepted values are "pad" and "interpolation".
        target_size : Optional[Tuple[int, int]], default None
            If given, the images will be resized to this size, instead of calculating a multiple of self.output_stride.
        pad_mode : str, default "replicate"
            Used if resize_mode == "pad". How to pad the input. Must be one of the values accepted by the 'mode' argument of torch.nn.functional.pad.
        pad_value : float, default 0.0
            Used if resize_mode == "pad" and pad_mode == "constant". The value to fill in the padded area.
        pad_two_side : bool, default True
            Used if resize_mode == "pad". If True, half of the padding goes to left/top and the rest to right/bottom. Otherwise, all the padding goes to the bottom right.
        interpolation_mode : str, default "bilinear"
            Used if resize_mode == "interpolation". How to interpolate the input. Must be one of the values accepted by the 'mode' argument of torch.nn.functional.interpolate.
        interpolation_align_corners : bool, default True
            Used if resize_mode == "interpolation". See 'align_corners' in https://pytorch.org/docs/stable/generated/torch.nn.functional.interpolate.html.

        Returns
        -------
        torch.Tensor
            A copy of the input images after applying all of the pre-processing steps.
        Union[InputPadder, InputScaler]
            An instance of InputPadder or InputScaler that was used to resize the images.
            Can be used to reverse the resizing operations.
        """
        bgr_add = bgr_val_as_tensor(
            bgr_add, reference_tensor=images, bgr_tensor_shape_position=-3
        )
        images = images + bgr_add
        bgr_mult = bgr_val_as_tensor(
            bgr_mult, reference_tensor=images, bgr_tensor_shape_position=-3
        )
        images *= bgr_mult
        if bgr_to_rgb:
            images = torch.flip(images, [-3])

        stride = self.output_stride
        if target_size is not None:
            stride = None

        if image_resizer is None:
            if resize_mode == "pad":
                image_resizer = InputPadder(
                    images.shape,
                    stride=stride,
                    size=target_size,
                    pad_mode=pad_mode,
                    two_side_pad=pad_two_side,
                    pad_value=pad_value,
                )
            elif resize_mode == "interpolation":
                image_resizer = InputScaler(
                    images.shape,
                    stride=stride,
                    size=target_size,
                    interpolation_mode=interpolation_mode,
                    interpolation_align_corners=interpolation_align_corners,
                )
            else:
                raise ValueError(
                    f"resize_mode must be one of (pad, interpolation). Found: {resize_mode}."
                )

        images = image_resizer.fill(images)
        images = images.contiguous()
        return images, image_resizer

    def postprocess_predictions(
        self,
        prediction: torch.Tensor,
        image_resizer: Optional[Union[InputPadder, InputScaler]],
        is_flow: bool,
    ) -> torch.Tensor:
        """Simple resizing post-processing. Just use image_resizer to revert the resizing operations.

        Parameters
        ----------
        prediction : torch.Tensor
            A tensor with at least 3 dimensions in this order: [..., C, H, W].
        image_resizer : Optional[Union[InputPadder, InputScaler]]
            An instance of InputPadder or InputScaler that will be used to reverse the resizing done to the inputs.
            Typically, this will be the instance returned by self.preprocess_images().
        is_flow : bool
            Indicates if prediction is an optical flow prediction of not.
            Only used if image_resizer is an instance of InputScaler, in which case the flow values need to be scaled.

        Returns
        -------
        torch.Tensor
            A copy of the prediction after reversing the resizing.
        """
        if isinstance(image_resizer, InputScaler):
            return image_resizer.unfill(prediction, is_flow=is_flow)
        else:
            return image_resizer.unfill(prediction)

    @abstractmethod
    def forward(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        """Forward the inputs through the network and produce the predictions.

        The method inputs can be anything, up to the implementation of the concrete model. However, the recommended input is
        to receive only dict[str, Any] as argument. This dict should contain everything required for one pass of the network
        (images, etc.). Arguments which do not change for each forward should be defined as arguments in the parser
        (see add_model_specific_args()).

        Parameters
        ----------
        args : Any
            Any arguments.
        kwargs : Any
            Any named arguments.

        Returns
        -------
        Dict[str, torch.Tensor]
            For compatibility with the framework, the output should be a dict containing at least the following keys:

            - 'flows': a 5D tensor BNCHW containing the predicted flows. The flows must be at the original scale, i.e.,
              their size is the same as the input images, and the flow magnitudes are scaled accordingly. Most networks only
              produce a single 2D optical flow per batch element, so the output shape will be B12HW. N can be larger than one
              if the network produces predictions for a larger temporal window.

            - 'occs': optional, and only included if the network also predicts occlusion masks. It is a 5D tensor following the
              same structure as 'flows'.

            - 'mbs': same as 'occs' but for occlusion masks.

            - 'confs': same as 'occs' but for flow confidence predictions.


            The keys above are used by other parts of PTLFlow (if available). The output can also include any other keys as
            well. For example, by default, the output of forward() will be passed to the loss function. So the output may
            include keys which are going to be used for computing the training loss.
        """
        pass

    def training_step(self, batch: Dict[str, Any], batch_idx: int) -> Dict[str, Any]:
        """Perform one step of the training.

        This function is called internally by Pytorch Lightning during training.

        Parameters
        ----------
        batch : Dict[str, Any]
            One batch of data, that is going to be given as input to the network.
        batch_idx : int
            The index of the current batch.

        Returns
        -------
        Dict[str, Any]
            A dict with two keys:

            - 'loss': torch.Tensor, containing the loss value. Required by Pytorch Lightning for the optimization step.

            - 'dataset_name': str, a string representing the name of the dataset from where this batch came from. Used only for
              logging purposes.
        """
        preds = self(batch)
        self.last_inputs = batch
        self.last_predictions = preds
        loss = self.loss_fn(preds, batch)
        metrics = self.train_metrics(preds, batch)
        if isinstance(loss, dict):
            for k, v in loss.items():
                metrics[f"train/{k}"] = v.item()
            loss = loss["loss"]
        else:
            metrics["train/loss"] = loss.item()
        self.log_dict(metrics, on_step=True, on_epoch=True)
        self.log(
            "epe", metrics["train/epe"], prog_bar=True, on_step=True, on_epoch=True
        )

        outputs = {"loss": loss, "dataset_name": batch["meta"]["dataset_name"]}
        return outputs

    def on_train_epoch_end(self) -> None:
        self.train_metrics.reset()

    def validation_step(
        self, batch: Dict[str, Any], batch_idx: int, dataloader_idx: int = 0
    ) -> Dict[str, Any]:
        """Perform one step of the validation.

        This function is called internally by Pytorch Lightning during validation.

        Parameters
        ----------
        batch : Dict[str, Any]
            One batch of data, that is going to be given as input to the network.
        batch_idx : int
            The index of the current batch.
        dataloader_idx : int, default 0
            When using multiple loaders, indicate from which loader this input is coming from.

        See Also
        --------
        ptlflow.utils.flow_metrics.FlowMetrics : class to manage and compute the optical flow metrics.
        """
        if len(self.val_metrics) <= dataloader_idx:
            self.val_metrics.append(
                FlowMetrics(prefix="val/").to(device=batch["flows"].device)
            )
            self.val_dataset_names.append(None)

        if self.warm_start:
            batch["prev_preds"] = self.prev_preds

        preds = self(batch)
        self.last_inputs = batch
        self.last_predictions = preds
        metrics = self.val_metrics[dataloader_idx](preds, batch)
        inputs_meta = batch.get("meta")
        train_val_metrics = self._split_train_val_metrics(metrics, inputs_meta)
        if (
            self.val_dataset_names[dataloader_idx] is None
            and inputs_meta is not None
            and inputs_meta.get("dataset_name") is not None
        ):
            name = inputs_meta.get("dataset_name")[0]
            if inputs_meta.get("split_name") is not None:
                name += f"-{inputs_meta.get('split_name')[0]}"
            self.val_dataset_names[dataloader_idx] = name
        self.log_dict(
            train_val_metrics,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
            add_dataloader_idx=False,
        )

        if self.warm_start:
            if "is_seq_start" in batch["meta"] and batch["meta"]["is_seq_start"][0]:
                self.prev_preds = None
            else:
                self.prev_preds = preds
                for k, v in self.prev_preds.items():
                    if isinstance(v, torch.Tensor):
                        self.prev_preds[k] = v.detach()

        return {"preds": preds, "metrics": metrics}

    def on_validation_epoch_end(self) -> None:
        for i in range(len(self.val_metrics)):
            metrics = self.val_metrics[i].compute()
            dset_name = self.val_dataset_names[i].lower()
            for name, val in metrics.items():
                main_metric = (
                    DATASET_MAIN_METRIC[dset_name]
                    if dset_name in DATASET_MAIN_METRIC
                    else "epe"
                )
                main_metric = f"val/{main_metric}"
                if name == main_metric:
                    self.log(dset_name, val, sync_dist=True, prog_bar=True)

                    if i == 0:
                        self.log("main_val_metric", val, sync_dist=True)
                        if not self.has_logged_main_val_metric_message:
                            logger.info(
                                "main_val_metric is tracking the metric {}/{}",
                                dset_name,
                                name,
                            )
                            self.has_logged_main_val_metric_message = True
            self.val_metrics[i].reset()

    def test_step(
        self, batch: Dict[str, Any], batch_idx: int, dataloader_idx: int = 0
    ) -> Dict[str, Any]:
        """Perform one step of the test.

        This function is called internally by Pytorch Lightning during testing.

        Parameters
        ----------
        batch : Dict[str, Any]
            One batch of data, that is going to be given as input to the network.
        batch_idx : int
            The index of the current batch.
        dataloader_idx : int, default 0
            When using multiple loaders, indicate from which loader this input is coming from.
        """
        if self.warm_start:
            batch["prev_preds"] = self.prev_preds

        preds = self(batch)
        self.last_inputs = batch
        self.last_predictions = preds

        if self.warm_start:
            if "is_seq_start" in batch["meta"] and batch["meta"]["is_seq_start"][0]:
                self.prev_preds = None
            else:
                self.prev_preds = preds
                for k, v in self.prev_preds.items():
                    if isinstance(v, torch.Tensor):
                        self.prev_preds[k] = v.detach()

        return preds

    def configure_optimizers(self) -> Dict[str, Any]:
        """Initialize the optimizers and LR schedulers.

        This function is called internally by Pytorch Lightning at the beginning of the training.

        Returns
        -------
        Dict[str, Any]
            A dict with two keys:
            - 'optimizer': an optimizer from PyTorch.
            - 'lr_scheduler': Dict['str', Any], a dict with the selected scheduler and its required arguments.
        """
        assert (
            self.loss_fn is not None
        ), f"Model {self.__class__.__name__} cannot be trained. It does not have a loss function."

        if self.trainer.max_steps is None or self.trainer.max_steps <= 0:
            device_divider = self.trainer.device_ids
            if isinstance(device_divider, list) or isinstance(device_divider, tuple):
                device_divider = len(device_divider)
            elif isinstance(device_divider, str):
                device_idx = [v for v in device_divider.split(",") if len(v) > 0]
                device_divider = len(device_idx)
            elif not isinstance(device_divider, int):
                device_divider = 1

            grad_batches_divider = (
                1
                if self.trainer.accumulate_grad_batches is None
                else self.trainer.accumulate_grad_batches
            )

            # Hack to get train_dataloader before it is set to self.trainer.train_dataloader
            steps_per_epoch = len(
                self.trainer.fit_loop._data_source.instance.train_dataloader()
            )
            max_steps = self.trainer.max_epochs * int(
                math.ceil(
                    float(steps_per_epoch) / device_divider / grad_batches_divider
                )
            )
            logger.info(
                "--trainer.max_steps is not provided. max_steps will be set as {} ({} epochs * {} steps per epoch / {} devices / {} accumulate grad steps)",
                max_steps,
                self.trainer.max_epochs,
                steps_per_epoch,
                device_divider,
                grad_batches_divider,
            )
        else:
            max_steps = self.trainer.max_steps

        if self.lr is None:
            self.lr = 1e-4
            logger.warning("--model.lr is not set. It will be set as {}", self.lr)
        if self.wdecay is None:
            self.wdecay = 1e-4
            logger.warning(
                "--model.wdecay is not set. It will be set as {}", self.wdecay
            )

        optimizer = optim.AdamW(self.parameters(), lr=self.lr, weight_decay=self.wdecay)

        lr_scheduler = optim.lr_scheduler.OneCycleLR(
            optimizer,
            self.lr,
            total_steps=max_steps,
            pct_start=0.05,
            cycle_momentum=False,
            anneal_strategy="linear",
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": {"scheduler": lr_scheduler, "interval": "step"},
        }

    def _split_train_val_metrics(
        self, metrics: Dict[str, float], inputs_meta: Optional[Dict[str, Any]] = None
    ) -> Dict[str, float]:
        """Split the val metrics into 'train', 'val', and 'full' categories.

        This is useful for producing metrics for different subsets without having to forward the same sample multiple times.
        When computing metrics for the full validation dataset, the metrics for the 'train' and 'val' subsets are also
        computed at the same time.

        Parameters
        ----------
        metrics : dict[str, float]
            The metrics computed after one validation step.
        inputs_meta : Optional[Dict[str, Any]], optional
            If this is not provided, then the metrics are not split. This is a dict with two keys:
            - 'dataset_name': str, a string representing the name of the dataset these metrics correspond to.
            - 'is_val': bool, indicating whether these metrics come from a sample in the 'val' split or not.

        Returns
        -------
        dict[str, float]
            The metrics after being split.

        See Also
        --------
        validation_step
        """
        dataset_name = None
        if inputs_meta is not None and inputs_meta.get("dataset_name") is not None:
            dataset_name = inputs_meta["dataset_name"][0].lower()

        log_metrics = {}
        has_train = False
        has_val = False
        for k, v in metrics.items():
            if k.startswith("val/"):
                k = k[4:]

            if inputs_meta is not None and inputs_meta.get("is_val") is not None:
                if inputs_meta["is_val"][0]:
                    split = "val"
                    has_val = True
                else:
                    split = "train"
                    has_train = True

                if dataset_name is not None:
                    log_metrics[f"val_{dataset_name}/{split}/{k}"] = v
                else:
                    log_metrics[f"val/{split}/{k}"] = v

        if has_train and has_val:
            for k, v in metrics.items():
                if k.startswith("val/"):
                    k = k[4:]

                if dataset_name is not None:
                    log_metrics[f"val_{dataset_name}/full/{k}"] = v
                else:
                    log_metrics[f"val/full/{k}"] = v

        return log_metrics



"""Operations to perform image augmentations for optical flow.

Some operations are adapted from the following sources:

- FlowNetPytorch: https://github.com/ClementPinard/FlowNetPytorch

- RAFT: https://github.com/princeton-vl/RAFT/

- flow-transforms-pytorch: https://github.com/hmorimitsu/flow-transforms-pytorch
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from collections.abc import KeysView
import random
from typing import Dict, Optional, Sequence, Tuple, Union

from einops import rearrange
import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms as tt


class Compose(object):
    """Similar to torchvision Compose. Applies a series of transforms from the input list in sequence."""

    def __init__(self, transforms_list: Sequence[object]) -> None:
        """Initialize Compose.

        Parameters
        ----------
        transforms_list : Sequence[object]
            A sequence of transforms to be applied.
        """
        self.transforms_list = [t for t in transforms_list if t is not None]

    def __call__(
        self, inputs: Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
            Each element of the dict is either a single 3D HWC image or a list of images.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        for t in self.transforms_list:
            inputs = t(inputs)
        return inputs


class ToTensor(object):
    """Converts a 4D numpy.ndarray or a list of 3D numpy.ndarrays into a 4D torch.Tensor.

    If an input is of type uint8, then it is converted to float and its values are divided by 255.
    """

    def __init__(
        self,
        fp16: bool = False,
        device: Union[str, torch.device] = "cpu",
        use_keys: Optional[Union[KeysView, Sequence[str]]] = None,
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
    ) -> None:
        """Initialize ToTensor.

        Parameters
        ----------
        fp16 : bool, default False
            If True, the tensors use have-precision floating point.
        device : Union[str, torch.device], default 'cpu'
            Name of the torch device where the tensors will be put in.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        self.dtype = torch.float16 if fp16 else torch.float32
        self.device = device
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys

    def __call__(
        self, inputs: Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, Union[np.ndarray, Sequence[np.ndarray]]]
            Each element of the dict is either a single 3D HWC image or a list of images.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for k in valid_keys:
            v = inputs[k]
            if isinstance(v, list) or isinstance(v, tuple):
                v = np.stack(v)
                if len(v.shape) == 3:
                    v = v[:, :, :, None]

            if len(v.shape) == 2:
                v = v[None, :, :, None]
            elif len(v.shape) == 3:
                v = v[None]

            if v.dtype == np.uint8:
                v = v.astype(np.float32) / 255.0
            v = v.transpose(0, 3, 1, 2)
            inputs[k] = torch.from_numpy(v).to(device=self.device, dtype=self.dtype)
        return inputs


class GenerateFBCheckFlowOcclusion(object):
    """Generate occlusion masks based on forward/backward flow consistency.

    In other words, a pixel p is considered occluded when \|Ff(p) + Fb(p + F(f))\|_2 > threshold,
    where Ff and Fb get the forward and backward flow vectors of a pixel.
    """

    def __init__(
        self,
        threshold: float = 0.0,
        forward_flow_key: str = "flows",
        backward_flow_key: str = "flows_b",
        forward_occlusion_key: str = "occs",
        backward_occlusion_key: str = "occs_b",
        compute_backward_occlusion: bool = True,
    ) -> None:
        """Initialize ColorJitter.

        Parameters
        ----------
        threshold : float, default 0.0
            A pixel is considered occluded if \|Ff(p) + Fb(p + F(f))\|_2 > threshold.
        forward_flow_key : str, default "flows"
            The name of the input dict entry that stores the forward flow.
        backward_flow_key : str, default "flows_b"
            The name of the input dict entry that stores the backward flow.
        forward_occlusion_key : str, default "occs"
            The name that will be added to the input dict to store the calculated occlusion masks for the forward flow.
        backward_occlusion_key : str, default "occs"
            The name that will be added to the input dict to store the calculated occlusion masks for the backward flow.
        compute_backward_occlusion : bool, default True
            If False, the occlusion mask is calculated only for the forward flow.
        """
        self.threshold = threshold
        self.forward_flow_key = forward_flow_key
        self.backward_flow_key = backward_flow_key
        self.forward_occlusion_key = forward_occlusion_key
        self.backward_occlusion_key = backward_occlusion_key
        self.compute_backward_occlusion = compute_backward_occlusion

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        assert self.forward_flow_key in inputs
        assert self.backward_flow_key in inputs
        flow_f = inputs[self.forward_flow_key]
        flow_b = inputs[self.backward_flow_key]
        b, c, h, w = flow_f.shape

        coords = self._coords_grid(b, h, w, dtype=flow_f.dtype, device=flow_f.device)
        flow_b_warped, oob_mask_f = self._bilinear_sampler(flow_b, coords + flow_f)
        diff_f = torch.norm(flow_f + flow_b_warped, p=2, dim=1, keepdim=True)
        inputs[self.forward_occlusion_key] = (
            (diff_f > self.threshold) | oob_mask_f
        ).float()

        if self.compute_backward_occlusion:
            flow_f_warped, oob_mask_b = self._bilinear_sampler(flow_f, coords + flow_b)
            diff_b = torch.norm(flow_b + flow_f_warped, p=2, dim=1, keepdim=True)
            inputs[self.backward_occlusion_key] = (
                (diff_b > self.threshold) | oob_mask_b
            ).float()

        return inputs

    def _bilinear_sampler(self, img, coords):
        # Code adapted from RAFT: https://github.com/princeton-vl/RAFT
        H, W = img.shape[-2:]
        xgrid, ygrid = coords.split([1, 1], dim=1)
        xgrid = 2 * xgrid / (W - 1) - 1
        ygrid = 2 * ygrid / (H - 1) - 1

        grid = torch.cat([xgrid, ygrid], dim=1)
        img = F.grid_sample(img, grid.permute(0, 2, 3, 1), align_corners=True)

        mask = (xgrid < -1) | (ygrid < -1) | (xgrid > 1) | (ygrid > 1)
        return img, mask

    def _coords_grid(self, batch, ht, wd, dtype, device):
        # Code adapted from RAFT: https://github.com/princeton-vl/RAFT
        coords = torch.meshgrid(
            torch.arange(ht, dtype=dtype, device=device),
            torch.arange(wd, dtype=dtype, device=device),
            indexing="ij",
        )
        coords = torch.stack(coords[::-1], dim=0)
        return coords[None].repeat(batch, 1, 1, 1)


class CenterCrop(object):
    """Applies center crop to the inputs."""

    def __init__(
        self,
        crop_size: Optional[Tuple[int, int]] = None,
        occlusion_keys: Union[KeysView, Sequence[str]] = ("occs", "occs_b"),
        valid_key: str = "valids",
        ignore_keys: Optional[Sequence[str]] = None,
    ) -> None:
        """Initialize RandomScaleAndCrop.

        Parameters
        ----------
        crop_size : Optional[Tuple[int, int]], optional
            If provided, crop the inputs to this size (h, w).
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        valid_keys : str, default 'valids'
            The name of the key in inputs that contains the binary mask indicating which pixels are valid.
            Only used when sparse=True.
        """
        self.crop_size = crop_size
        self.occlusion_keys = list(occlusion_keys)
        self.valid_key = valid_key
        self.ignore_keys = ignore_keys

    def __call__(  # noqa: C901
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.

        Raises
        ------
        NotImplementedError
            If trying to use time scale.
        """
        h, w = inputs[self.valid_key].shape[2:4]
        y_crop = (h - self.crop_size[0]) // 2
        x_crop = (w - self.crop_size[1]) // 2

        for k, v in inputs.items():
            if k not in self.ignore_keys:
                v = v[
                    :,
                    :,
                    y_crop : y_crop + self.crop_size[0],
                    x_crop : x_crop + self.crop_size[1],
                ]
            inputs[k] = v

        # Update occlusion masks for out-of-bounds flows
        for k, v in inputs.items():
            if self.ignore_keys is None or k not in self.ignore_keys:
                try:
                    i = self.occlusion_keys.index(k)
                    inputs[k] = _update_oob_flows(v, inputs[self.flow_keys[i]])
                except ValueError:
                    pass
        return inputs


class ColorJitter(tt.ColorJitter):
    """Randomly apply color transformations only to the images.

    If asymmetric_prob == 0, then the same transform is applied on all the images, otherwise, the transform for each image
    is randomly sampled independently.

    This is basically a wrapper for torchvision.transforms.ColorJitter.
    """

    def __init__(
        self,
        brightness: Union[float, Tuple[float, float]] = 0.0,
        contrast: Union[float, Tuple[float, float]] = 0.0,
        saturation: Union[float, Tuple[float, float]] = 0.0,
        hue: Union[float, Tuple[float, float]] = 0.0,
        asymmetric_prob: float = 0.0,
        use_keys: Optional[Union[KeysView, Sequence[str]]] = ("images",),
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
    ) -> None:
        """Initialize ColorJitter.

        Parameters
        ----------
        brightness : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random brightness value.
        contrast : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random contrast value.
        saturation : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random saturation value.
        hue : Union[float, Tuple[float, float]], default 0.0
            The range to sample the random hue value.
        asymmetric_prob : float, default 0.0
            Chance to apply an asymmetric transform, in which the parameters for transforming each image are sampled
            independently.
        use_keys : Optional[Union[KeysView, Sequence[str]]], default ['images']
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        super().__init__(
            brightness=brightness, contrast=contrast, saturation=saturation, hue=hue
        )
        self.asymmetric_prob = asymmetric_prob
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for k in valid_keys:
            v = inputs[k]
            if random.random() < self.asymmetric_prob:
                for i in range(len(v)):
                    inputs[k][i] = super().__call__(v[i])
            else:
                inputs[k] = super().__call__(v)
        return inputs


class GaussianNoise(object):
    """Applies random gaussian noise on the images."""

    def __init__(
        self,
        stdev: float = 0.0,
        use_keys: Optional[Union[KeysView, Sequence[str]]] = ("images",),
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
    ) -> None:
        """Initialize GaussianNoise.

        Parameters
        ----------
        stdev : float, default 0.0
            The maximum standard deviation of the gaussian noise.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        self.stdev = stdev
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for k in valid_keys:
            v = inputs[k]
            std = random.uniform(0.0, self.stdev)
            inputs[k] = (
                v + std * torch.randn(*v.shape, dtype=v.dtype, device=v.device)
            ).clamp(0.0, 1.0)
        return inputs


class RandomPatchEraser(object):
    """Randomly covers a rectangular patch on the second image with noise, to simulate a pseudo-occlusion.

    The noise_type may be the mean or random. This transform erases patches ONLY FROM THE SECOND IMAGE.
    """

    def __init__(
        self,
        erase_prob: float = 0.0,
        num_patches: Union[int, Tuple[int, int]] = 1,
        patch_size: Union[Tuple[int, int], Tuple[int, int, int, int]] = (0, 0),
        noise_type: str = "mean",
        use_keys: Optional[Union[KeysView, Sequence[str]]] = ("images",),
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
    ) -> None:
        """Initialize RandomPatchEraser.

        Parameters
        ----------
        erase_prob : float, default 0.0
            Probability of applying the transformation.
        num_patches : Union[int, Tuple[int, int]], default 1
            Number of occlusion patches to generate. If it is a tuple, the number will be uniformly sampled from the interval.
        patch_size : Union[Tuple[int, int], Tuple[int, int, int, int]], default (0, 0)
            Range of the size of the occlusion patches. If it is a tuple with 2 elements, then both sides are sampled from the
            same interval. Otherwise, different intervals can be specified for each side as (hmin, hmax, wmin, wmax).
        noise_type : str, default 'mean'
            How to fill the occlusion patch. It can be either with the image 'mean' or with random 'noise'.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        """
        self.erase_prob = erase_prob
        self.noise_type = noise_type
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys
        self.num_patches = num_patches
        if not (isinstance(num_patches, tuple) or isinstance(num_patches, list)):
            self.num_patches = (num_patches, num_patches)
        self.patch_size = patch_size
        if len(patch_size) == 2:
            self.patch_size = (
                patch_size[0],
                patch_size[1],
                patch_size[0],
                patch_size[1],
            )

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        if random.random() < self.erase_prob:
            valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
            for k in valid_keys:
                im2 = inputs[k][1]
                c, h, w = im2.shape
                if self.noise_type == "mean":
                    mean_color = im2.reshape(c, -1).mean(dim=1)
                for _ in range(
                    random.randint(self.num_patches[0], self.num_patches[1])
                ):
                    hp = random.randint(self.patch_size[0], self.patch_size[1])
                    wp = random.randint(self.patch_size[2], self.patch_size[3])
                    if hp > 0 and wp > 0:
                        yp = random.randint(0, h - hp)
                        xp = random.randint(0, w - wp)
                        if self.noise_type == "mean":
                            noise = mean_color[:, None, None].repeat(1, hp, wp)
                        else:
                            im_min = im2.min()
                            im_max = im2.max()
                            im_inter = im_max - im_min
                            noise = (
                                im_inter
                                * torch.rand(
                                    c, hp, wp, dtype=im2.dtype, device=im2.device
                                )
                                + im_min
                            )
                        im2[:, yp : yp + hp, xp : xp + wp] = noise
        return inputs


class RandomFlip(object):
    """Randomly horizontally and vertically flips the inputs.

    If asymmetric_prob > 0, then each input of the sequence may be flipped differently.
    """

    def __init__(
        self,
        hflip_prob: float = 0.0,
        vflip_prob: float = 0.0,
        asymmetric_prob: float = 0.0,
        use_keys: Optional[Union[KeysView, Sequence[str]]] = None,
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
        image_keys: Union[KeysView, Sequence[str]] = ("images",),
        flow_keys: Union[KeysView, Sequence[str]] = ("flows", "flows_b"),
    ) -> None:
        """Initialize RandomFlip.

        Parameters
        ----------
        hflip_prob : float, default 0.0
            Probability of applying a horizontal flip.
        vflip_prob : float, default 0.0
            Probability of applying a vertical flip.
        asymmetric_prob : float, default 0.0
            Chance to apply an asymmetric transform, in which the parameters for transforming each image are sampled
            independently.
        use_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed,
            except the keys that are listed in ignore_keys.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]], optional
            If use_keys is None, the these keys are NOT transformed by this operation.
        image_keys : Union[KeysView, Sequence[str]], ['images']
            Indicate which of the input keys correspond to image tensors.
        flow_keys : Union[KeysView, Sequence[str]], ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        """
        self.flip_probs = [hflip_prob, vflip_prob]
        self.asymmetric_prob = asymmetric_prob
        self.use_keys = use_keys
        self.ignore_keys = ignore_keys
        self.image_keys = list(image_keys)
        self.flow_keys = list(flow_keys)

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        valid_keys = _get_valid_keys(inputs.keys(), self.use_keys, self.ignore_keys)
        for iorient in range(2):
            if self.asymmetric_prob < 1e-5:
                if random.random() < self.flip_probs[iorient]:
                    inputs = self._flip_inputs(inputs, iorient == 0, valid_keys)
            else:
                is_flips = [
                    random.random() < self.flip_probs[iorient]
                    for _ in range(inputs[self.image_keys[0]].shape[0])
                ]
                for i in range(inputs[self.flow_keys[0]].shape[0]):
                    if is_flips[i]:
                        inputs = self._flip_inputs(
                            inputs, iorient == 0, valid_keys, ibatch=i
                        )
                    if is_flips[i] != is_flips[i + 1]:
                        for fk in self.flow_keys:
                            inputs[fk][i] = self._mirror_flow(
                                inputs[fk][i], iorient == 0
                            )
                if is_flips[-1]:
                    for ik in self.image_keys:
                        inputs = self._flip_inputs(
                            inputs, iorient == 0, inputs_keys=[ik], ibatch=-1
                        )

        return inputs

    def _flip_inputs(
        self,
        inputs: Dict[str, torch.Tensor],
        is_hflip: bool,
        valid_keys: Optional[Sequence[str]] = None,
        ibatch: Optional[int] = None,
    ) -> Dict[str, torch.Tensor]:
        """Flips all inputs horizontally or vertically.

        This function properly adjust the flow values after the flipping.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be flipped. Each element is a 4D tensor NCHW.
        is_hflip : bool
            If True, performs a horizontal flip, otherwise, performs a vertical flip.
        valid_keys : Optional[Sequence[str]], optional
            If it is not None, then only elements with these keys will be transformed. Otherwise, all elements are transformed.
        ibatch : Optional[int], optional
            If ibatch is specified, then only one element of the batch is flipped.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs flipped by this operation.
        """
        if is_hflip:
            iinp = 3
            iflow = 0
        else:
            iinp = 2
            iflow = 1

        if valid_keys is None:
            valid_keys = list(inputs.keys())
        for k in valid_keys:
            if ibatch is None:
                inputs[k] = torch.flip(inputs[k], [iinp])
                if "flows" in k:
                    inputs[k][:, iflow] *= -1
            else:
                inputs[k][ibatch] = torch.flip(inputs[k][ibatch], [iinp - 1])
                if "flows" in k:
                    inputs[k][ibatch, iflow] *= -1
        return inputs

    def _mirror_flow(self, flow: torch.Tensor, is_hflip: bool) -> torch.Tensor:
        """Reflects the flow along the center line of the image.

        This function is used when an asymmetric flip happens (one image flips, but the next does not, or vice-versa).

        Parameters
        ----------
        flow : torch.Tensor
            A 3D tensor CHW.
        is_hflip : bool
            If True, performs a horizontal flip, otherwise, performs a vertical flip.

        Returns
        -------
        torch.Tensor
            The mirrored flow.
        """
        grid = torch.meshgrid(
            torch.arange(flow.shape[1]), torch.arange(flow.shape[2]), indexing="ij"
        )
        grid = torch.stack(grid[::-1]).float()
        if is_hflip:
            mean_coord = (flow.shape[2] - 1) / 2.0
            flow[0] = 2 * (mean_coord - grid[0]) - flow[0]
        else:
            mean_coord = (flow.shape[1] - 1) / 2.0
            flow[1] = 2 * (mean_coord - grid[1]) - flow[1]
        return flow


class RandomScaleAndCrop(object):
    """Applies first random scale and then random crop to the inputs.

    The scale is adjusted so that it is not smaller than the crop size.

    The scale calculation is composed of 2 main stages:

    1. A random major scale is sampled. The major scale defines the global scale applied to
       all images and dimensions. The major scale is calculated as:

       ms = 2 ** random.uniform(major_scale[0], major_scale[1])

    2. A random minor space scale is sampled. The space scale dictates the variation
       in scale applied to the width and height of each image. The space scale is
       calculated as:

       ssh = 2 ** random.uniform(space_scale[0], space_scale[1])

       ssw = 2 ** random.uniform(space_scale[2], space_scale[3]).

       If len(space_scale) == 2, then ssw also uses space_scale[0] and space_scale[1].

    The final scale applied to all inputs is:

        scale_height = ms * ssh

        scale_width = ms * ssw

    If time_scale is provided, then a third scale is sampled before computing the final scale.
    The time_scale is sampled independently for each element of a sequence. This allows,
    for example, for the first image have a different scale then the second one.
    The time scales tsh and tsw are calculated as the space scales ssh and ssw.
    With time scales, the final scales are calculated as:

        scale_height_time_t = ms * ssh * tsh_t

        scale_width_time_t = ms * ssw * tsw_t
    """

    def __init__(
        self,
        crop_size: Optional[Tuple[int, int]] = None,
        major_scale: Tuple[float, float] = (0.0, 0.0),
        space_scale: Union[Tuple[float, float], Tuple[float, float, float, float]] = (
            0.0,
            0.0,
        ),
        time_scale: Union[Tuple[float, float], Tuple[float, float, float, float]] = (
            0.0,
            0.0,
        ),
        binary_keys: Union[KeysView, Sequence[str]] = (
            "mbs",
            "occs",
            "valids",
            "mbs_b",
            "occs_b",
            "valids_b",
        ),
        flow_keys: Union[KeysView, Sequence[str]] = ("flows", "flows_b"),
        occlusion_keys: Union[KeysView, Sequence[str]] = ("occs", "occs_b"),
        sparse: bool = False,
        valid_key: str = "valids",
    ) -> None:
        """Initialize RandomScaleAndCrop.

        Parameters
        ----------
        crop_size : Optional[Tuple[int, int]], optional
            If provided, crop the inputs to this size (h, w).
        major_scale : Tuple[float, float], default (0.0, 0.0)
            The range of the major scale. See the class description for more details.
        space_scale : Union[Tuple[float, float], Tuple[float, float, float, float]], default (0.0, 0.0)
            The range of the minor scale. See the class description for more details.
        time_scale : Union[Tuple[float, float], Tuple[float, float, float, float]], default (0.0, 0.0)
            NOTE: Currently not implemented. The range of the time scale. See the class description for more details.
        binary_keys : Union[KeysView, Sequence[str]], default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
            Indicate which of the input keys correspond to binary tensors.
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        sparse : bool, default False
            If True, only values at valid positions (indicated by the mask in inputs[valid_key]) will be kept when
            resizing binary and flow inputs. Requires valid_key to exist as a key in inputs.
        valid_keys : str, default 'valids'
            The name of the key in inputs that contains the binary mask indicating which pixels are valid.
            Only used when sparse=True.
        """
        self.crop_size = crop_size
        self.major_scale = major_scale
        if len(major_scale) == 2:
            self.major_scale = (
                major_scale[0],
                major_scale[1],
                major_scale[0],
                major_scale[1],
            )
        self.space_scale = space_scale
        if len(space_scale) == 2:
            self.space_scale = (
                space_scale[0],
                space_scale[1],
                space_scale[0],
                space_scale[1],
            )
        self.time_scale = time_scale
        if len(time_scale) == 2:
            self.time_scale = (
                time_scale[0],
                time_scale[1],
                time_scale[0],
                time_scale[1],
            )

        self.use_time_scale = (
            abs(self.time_scale[1] - self.time_scale[0]) > 1e-5
            and abs(self.time_scale[3] - self.time_scale[2]) > 1e-5
        )
        self.binary_keys = list(binary_keys)
        self.flow_keys = list(flow_keys)
        self.occlusion_keys = list(occlusion_keys)
        self.sparse = sparse
        self.valid_key = valid_key

    def __call__(  # noqa: C901
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.

        Raises
        ------
        NotImplementedError
            If trying to use time scale.
        """
        h, w = inputs[self.flow_keys[0]].shape[2:4]
        major_scale = 2 ** random.uniform(self.major_scale[0], self.major_scale[1])
        space_scales = (
            2 ** random.uniform(self.space_scale[0], self.space_scale[1]),
            2 ** random.uniform(self.space_scale[2], self.space_scale[3]),
        )
        if self.use_time_scale:
            raise NotImplementedError("time_scale is currently not supported")
        else:
            min_size = self.crop_size
            if min_size is None:
                min_size = (1, 1)
            scaled_size = (
                max(min_size[0], int(h * major_scale * space_scales[0])),
                max(min_size[1], int(w * major_scale * space_scales[1])),
            )
            if self.crop_size is not None:
                y_crop = random.randint(0, scaled_size[0] - self.crop_size[0])
                x_crop = random.randint(0, scaled_size[1] - self.crop_size[1])

            inputs = _resize(
                inputs,
                scaled_size,
                self.binary_keys,
                self.flow_keys,
                self.sparse,
                self.valid_key,
            )
            if self.crop_size is not None:
                for k, v in inputs.items():
                    v = v[
                        :,
                        :,
                        y_crop : y_crop + self.crop_size[0],
                        x_crop : x_crop + self.crop_size[1],
                    ]
                    inputs[k] = v

            # Update occlusion masks for out-of-bounds flows
            for k, v in inputs.items():
                try:
                    i = self.occlusion_keys.index(k)
                    inputs[k] = _update_oob_flows(v, inputs[self.flow_keys[i]])
                except ValueError:
                    pass
        return inputs


class RandomTranslate(object):
    """Creates a translation between images by applying a random alternated crop on the sequence of inputs.

    A translation value t is randomly selected first. Then, the first image is cropped by a box translated by t.
    The second image will be cropped by a reversed translation -t. The third will be cropped by t again, and so on...
    """

    def __init__(
        self,
        translation: Union[int, Tuple[int, int]] = 0,
        flow_keys: Union[KeysView, Sequence[str]] = ("flows", "flows_b"),
        occlusion_keys: Union[KeysView, Sequence[str]] = ("occs", "occs_b"),
    ) -> None:
        """Initialize RandomTranslate.

        Parameters
        ----------
        translation : Union[int, Tuple[int, int]], default 0
            Maximum translation (in pixels) to be applied to the inputs. If a tuple, it corresponds to the maximum in the
            (y, x) axes.
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        """
        self.translation = translation
        if not isinstance(translation, tuple) or isinstance(translation, list):
            self.translation = (translation, translation)
        self.flow_keys = flow_keys
        self.occlusion_keys = occlusion_keys

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        _, _, h, w = inputs[self.flow_keys[0]].shape
        th, tw = self.translation
        tw = random.randint(-tw, tw)
        th = random.randint(-th, th)
        if tw == 0 and th == 0:
            return inputs

        trans_inputs = {
            k: torch.empty_like(v[:, :, : h - abs(th), : w - abs(tw)])
            for k, v in inputs.items()
        }

        # Translate: 0: even indexed inputs, 1: odd indexed inputs
        for t in range(2):
            if t == 0:
                ftw = tw
                fth = th
            else:
                ftw = -tw
                fth = -th
            x1, x2 = max(0, ftw), min(w + ftw, w)
            y1, y2 = max(0, fth), min(h + fth, h)
            for k, v in inputs.items():
                trans_inputs[k][t::2] = v[t::2, :, y1:y2, x1:x2]
                if k in self.flow_keys:
                    trans_inputs[k][t::2, 0] += ftw
                    trans_inputs[k][t::2, 1] += fth

        # Update occlusion masks for out-of-bounds flows
        for k, v in trans_inputs.items():
            try:
                i = self.occlusion_keys.index(k)
                trans_inputs[k] = _update_oob_flows(v, trans_inputs[self.flow_keys[i]])
            except ValueError:
                pass

        return trans_inputs


class RandomRotate(object):
    """Applies random rotation to the inputs.

    The inputs are rotated around the center of the image. First all inputs are rotated by the same random major `angle`.
    Then, another random angle a is sampled according to `diff_angle`. The first image will be rotated by a. The second image
    will be rotated by a reversed angle -a. The third will be rotated by a again, and so on...
    """

    def __init__(
        self,
        angle: float = 0.0,
        diff_angle: float = 0.0,
        flow_keys: Union[KeysView, Sequence[str]] = ("flows", "flows_b"),
        occlusion_keys: Union[KeysView, Sequence[str]] = ("occs", "occs_b"),
        valid_keys: Union[KeysView, Sequence[str]] = ("valids", "valids_b"),
        binary_keys: Union[KeysView, Sequence[str]] = (
            "mbs",
            "occs",
            "valids",
            "mbs_b",
            "occs_b",
            "valids_b",
        ),
        sparse: bool = False,
    ) -> None:
        """Initialize RandomRotate.

        Parameters
        ----------
        angle : float, default 0.0
            The maximum absolute value to sample the major angle from.
        diff_angle : float, default 0.0
            The maximum absolute value to sample the angle difference between consecutive images.
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        occlusion_keys : Union[KeysView, Sequence[str]], default ['occs', 'occs_b']
            Indicate which of the input keys correspond to occlusion mask tensors.
        valid_keys : Union[KeysView, Sequence[str]], default ['valids', 'valids_b']
            Indicate which of the input keys correspond to valid mask tensors.
        binary_keys : Union[KeysView, Sequence[str]], default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
            Indicate which of the input keys correspond to binary tensors.
        sparse : bool, default False
            If True, all binary inputs and flows are rotated with nearest grid_sample, instead of bilinear.
        """
        self.angle = angle
        self.diff_angle = diff_angle
        self.flow_keys = flow_keys
        self.occlusion_keys = occlusion_keys
        self.valid_keys = valid_keys
        self.binary_keys = binary_keys
        self.sparse = sparse

    def __call__(  # noqa: C901
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        major_angle = random.uniform(-self.angle, self.angle)
        inter_angle = random.uniform(-self.diff_angle, self.diff_angle)

        input_tensor = inputs[self.flow_keys[0]]
        b, _, h, w = input_tensor.shape

        def generate_rotation_grid(
            rot_angle: float, batch_size: int, dtype: torch.dtype, device: torch.device
        ) -> torch.Tensor:
            vy, vx = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
            vx = vx.type(dtype)
            vy = vy.type(dtype)
            vx = vx.to(device)
            vy = vy.to(device)
            vx -= (w - 1.0) / 2.0
            vy -= (h - 1.0) / 2.0
            angle_rad = rot_angle * 2 * np.pi / 360
            rotx = np.cos(angle_rad) * vx - np.sin(angle_rad) * vy
            roty = np.sin(angle_rad) * vx + np.cos(angle_rad) * vy
            rotx = rotx / ((w - 1) / 2)
            roty = roty / ((h - 1) / 2)
            rot_grid = torch.stack((rotx, roty), dim=2)[None]
            rot_grid = rot_grid.repeat(batch_size, 1, 1, 1)
            return rot_grid

        def generate_rotation_matrix(
            rot_angle: float, batch_size: int, dtype: torch.dtype, device: torch.device
        ) -> torch.Tensor:
            vx, vy = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
            vx = vx.type(dtype)
            vy = vy.type(dtype)
            vx = vx.to(device)
            vy = vy.to(device)
            rotx = (vx - h / 2.0) * (rot_angle * np.pi / 180.0)
            roty = -(vy - w / 2.0) * (rot_angle * np.pi / 180.0)
            rot_mat = torch.stack((rotx, roty), dim=0)[None]
            rot_mat = rot_mat.repeat(batch_size, 1, 1, 1)
            return rot_mat

        def rotate_flow(flow: torch.Tensor, rot_angle: float) -> torch.Tensor:
            angle_rad = rot_angle * 2 * np.pi / 360
            rot_flow = flow.clone()
            rot_flow[:, 0] = (
                np.cos(angle_rad) * flow[:, 0] + np.sin(angle_rad) * flow[:, 1]
            )
            rot_flow[:, 1] = (
                -np.sin(angle_rad) * flow[:, 0] + np.cos(angle_rad) * flow[:, 1]
            )
            return rot_flow

        rot_mat = generate_rotation_matrix(
            inter_angle, b // 2 + 1, input_tensor.dtype, input_tensor.device
        )
        for t in range(2):
            if t == 0:
                inangle = -inter_angle
                rmat = rot_mat
            else:
                inangle = inter_angle
                rmat = -rot_mat
            angle = major_angle + inangle / 2
            num_flows = input_tensor[t::2].shape[0]
            num_images = num_flows + 1
            rot_grid = generate_rotation_grid(
                angle, num_images, input_tensor.dtype, input_tensor.device
            )
            for k, v in inputs.items():
                if k in self.flow_keys:
                    v[t::2] += rmat[:num_flows]

                if k in self.binary_keys:
                    v[t::2] = F.grid_sample(
                        v[t::2], rot_grid[: v[t::2].shape[0]], mode="nearest"
                    )
                else:
                    if k in self.flow_keys:
                        if self.sparse:
                            v[t::2] = F.grid_sample(
                                v[t::2], rot_grid[: v[t::2].shape[0]], mode="nearest"
                            )
                        else:
                            v[t::2] = F.grid_sample(
                                v[t::2],
                                rot_grid[: v[t::2].shape[0]],
                                mode="bilinear",
                                align_corners=True,
                            )
                        v[t::2] = rotate_flow(v[t::2], angle)
                    else:
                        v[t::2] = F.grid_sample(
                            v[t::2],
                            rot_grid[: v[t::2].shape[0]],
                            mode="bilinear",
                            align_corners=True,
                        )

                if k in self.flow_keys:
                    v[t::2] = rotate_flow(v[t::2], angle)

                inputs[k] = v

        # Update occlusion masks for out-of-bounds flows
        for k, v in inputs.items():
            try:
                i = self.occlusion_keys.index(k)
                v = _update_oob_flows(v, inputs[self.flow_keys[i]])
                inputs[k] = v
            except ValueError:
                pass

        return inputs


class Resize(object):
    """Resize the image to a given size or scale.

    Size is checked first, if any of its values is zero, then scale is used.
    """

    def __init__(
        self,
        size: Tuple[int, int] = (0, 0),
        scale: float = 1.0,
        binary_keys: Union[KeysView, Sequence[str]] = (
            "mbs",
            "occs",
            "valids",
            "mbs_b",
            "occs_b",
            "valids_b",
        ),
        flow_keys: Union[KeysView, Sequence[str]] = ("flows", "flows_b"),
        sparse: bool = False,
        valid_key: str = "valids",
        ignore_keys: Optional[Union[KeysView, Sequence[str]]] = None,
    ) -> None:
        """Initialize Resize.

        Parameters
        ----------
        size : Tuple[int, int], default (0, 0)
            The target size to resize the inputs. If it is zeros, then the scale will be used instead.
        scale : float, default 1.0
            The scale factor to resize the images. Only used if size is zeros.
        binary_keys : Union[KeysView, Sequence[str]], default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
            Indicate which of the input keys correspond to binary tensors.
            [description], by default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
        flow_keys : Union[KeysView, Sequence[str]], default ['flows', 'flows_b']
            Indicate which of the input keys correspond to optical flow tensors.
        sparse : bool, default False
            If True, only values at valid positions (indicated by the mask in inputs[valid_key]) will be kept when
            resizing binary and flow inputs. Requires valid_key to exist as a key in inputs.
        valid_keys : str, default 'valids'
            The name of the key in inputs that contains the binary mask indicating which pixels are valid.
            Only used when sparse=True.
        ignore_keys : Optional[Union[KeysView, Sequence[str]]]
            If not None, remove these keys from the inputs_keys.
        """
        self.size = size
        self.scale = scale
        self.binary_keys = list(binary_keys)
        self.flow_keys = list(flow_keys)
        self.sparse = sparse
        self.valid_key = valid_key
        self.ignore_keys = ignore_keys

    def __call__(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform the transformation on the inputs.

        Parameters
        ----------
        inputs : Dict[str, torch.Tensor]
            Elements to be transformed. Each element is a 4D tensor NCHW.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs transformed by this operation.
        """
        h, w = inputs[list(inputs.keys())[0]].shape[2:4]
        if self.size is None or self.size[0] < 1 or self.size[1] < 1:
            self.size = (int(self.scale * h), int(self.scale * w))
        if self.size[0] != h or self.size[1] != w:
            inputs = _resize(
                inputs,
                self.size,
                self.binary_keys,
                self.flow_keys,
                self.sparse,
                self.valid_key,
                ignore_keys=self.ignore_keys,
            )
        return inputs


def _get_valid_keys(
    inputs_keys: Union[KeysView, Sequence[str]],
    use_keys: Optional[Union[KeysView, Sequence[str]]],
    ignore_keys: Optional[Union[KeysView, Sequence[str]]],
) -> Union[KeysView, Sequence[str]]:
    """Get only the valid keys from the input.

    Basically, it return use_keys, if not None. Otherwise, return inputs_keys after removing the keys which are in ignore_keys.

    Parameters
    ----------
    inputs_keys : Union[KeysView, Sequence[str]]
        All the keys available as input.
    use_keys : Optional[Union[KeysView, Sequence[str]]]
        If not None, then just use these keys.
    ignore_keys : Optional[Union[KeysView, Sequence[str]]]
        If not None, remove these keys from the inputs_keys.

    Returns
    -------
    Union[KeysView, Sequence[str]]
        The keys remaining after the validity checks.
    """
    if use_keys is not None:
        return use_keys
    if ignore_keys is None:
        return inputs_keys
    return [k for k in inputs_keys if k not in ignore_keys]


def _resize(
    inputs: Dict[str, torch.Tensor],
    target_size: Tuple[int, int],
    binary_keys: Union[KeysView, Sequence[str]],
    flow_keys: Union[KeysView, Sequence[str]],
    sparse: bool,
    valid_key: str,
    ignore_keys: Optional[Sequence[str]] = None,
):
    """Resize inputs to a target size. Set sparse=True when the valid mask has holes.
    This ensures that the resized valid mask does not interpolate the valid positions.

    Parameters
    ----------
    inputs : Dict[str, torch.Tensor]
        Elements to be transformed. Each element is a 4D tensor NCHW.
    target_size : Tuple[int, int]
        Target (height, width) sizes.
    binary_keys : Union[KeysView, Sequence[str]]
        Indicate which of the input keys correspond to binary tensors.
        [description], by default ['mbs', 'occs', 'valids', 'mbs_b', 'occs_b', 'valids_b']
    flow_keys : Union[KeysView, Sequence[str]]
        Indicate which of the input keys correspond to optical flow tensors.
    sparse : bool
        If True, only values at valid positions (indicated by the mask in inputs[valid_key]) will be kept when
        resizing binary and flow inputs. Requires valid_key to exist as a key in inputs.
    valid_keys : str
        The name of the key in inputs that contains the binary mask indicating which pixels are valid.
        Only used when sparse=True.

    Returns
    -------
    torch.Tensor
        The updated occlusion masks. Flows which went out-of-bounds are marked as occluded.
    """
    if sparse:
        assert (
            valid_key in inputs
        ), f"sparse is True, but valid_key({valid_key}) is not in inputs"
        valids = inputs[valid_key]
        n, k, h, w = valids.shape
        hs, ws = target_size
        scale_factor = torch.tensor(
            [float(ws) / w, float(hs) / h], device=valids.device
        )
        valids_flat = rearrange(valids, "n k h w -> n (k h w)")
        xy_scaled_list = []
        inbounds_list = []
        valids_out = torch.zeros(n, k, hs, ws, dtype=torch.float, device=valids.device)
        for i, vflat in enumerate(valids_flat):
            coords = torch.meshgrid(
                torch.arange(w, device=valids.device),
                torch.arange(h, device=valids.device),
                indexing="ij",
            )
            coords = torch.stack(coords, dim=-1)
            coords = rearrange(coords, "h w k -> (w h) k")

            coords_valid = coords[vflat >= 1]

            coords_scaled = coords_valid * scale_factor

            x_scaled = torch.round(coords_scaled[:, 0]).long()
            y_scaled = torch.round(coords_scaled[:, 1]).long()
            inbounds = (
                (x_scaled > 0) & (x_scaled < ws) & (y_scaled > 0) & (y_scaled < hs)
            )
            inbounds_list.append(inbounds)
            x_scaled = x_scaled[inbounds]
            y_scaled = y_scaled[inbounds]
            xy_scaled_list.append((x_scaled, y_scaled))

            valids_out[i, 0, y_scaled, x_scaled] = 1

        inputs[valid_key] = valids_out

        for k, v in inputs.items():
            if k != valid_key and (ignore_keys is None or k not in ignore_keys):
                if k in binary_keys or k in flow_keys:
                    v_out = torch.zeros(
                        v.shape[0], v.shape[1], hs, ws, dtype=v.dtype, device=v.device
                    )
                    for i, v_one in enumerate(v):
                        v_flat = rearrange(v_one, "k h w -> (h w) k")
                        v_valid = v_flat[valids_flat[i] >= 1]
                        if k in flow_keys:
                            v_valid = v_valid * scale_factor
                        v_valid = v_valid[inbounds_list[i]]
                        v_valid = rearrange(v_valid, "n k -> k n")
                        v_out[i, :, xy_scaled_list[i][1], xy_scaled_list[i][0]] = (
                            v_valid
                        )
                    v = v_out
                else:
                    v = F.interpolate(
                        v, size=target_size, mode="bilinear", align_corners=True
                    )
            inputs[k] = v
    else:
        for k, v in inputs.items():
            if ignore_keys is None or k not in ignore_keys:
                h, w = v.shape[-2:]
                if k in binary_keys:
                    v = F.interpolate(v, size=target_size, mode="nearest")
                else:
                    v = F.interpolate(
                        v, size=target_size, mode="bilinear", align_corners=True
                    )

                if k in flow_keys:
                    scale_mult = torch.tensor(
                        [float(target_size[1]) / w, float(target_size[0]) / h],
                        device=v.device,
                    )[None, :, None, None]
                    v = v * scale_mult

            inputs[k] = v

    return inputs


def _update_oob_flows(occs: torch.Tensor, flows: torch.Tensor) -> torch.Tensor:
    """Update occlusion maps to include flow which went out-of-bounds.

    Parameters
    ----------
    occs : torch.Tensor
        A 4D tensor NCHW of occlusion masks.
    flows : torch.Tensor
        A 4D tensor NCHW of optical flows.

    Returns
    -------
    torch.Tensor
        The updated occlusion masks. Flows which went out-of-bounds are marked as occluded.
    """
    grid = torch.meshgrid(
        torch.arange(flows.shape[2], dtype=flows.dtype, device=flows.device),
        torch.arange(flows.shape[3], dtype=flows.dtype, device=flows.device),
        indexing="ij",
    )
    grid = torch.stack(grid[::-1]).float()[None].repeat(flows.shape[0], 1, 1, 1)
    coords = flows + grid
    oob_occs = coords < 0
    oob_occs[:, 0] |= coords[:, 0] > flows.shape[3]
    oob_occs[:, 1] |= coords[:, 1] > flows.shape[2]
    oob_occs = oob_occs.max(dim=1, keepdim=True)[0].to(
        dtype=occs.dtype, device=occs.device
    )
    occs = torch.max(torch.stack([occs, oob_occs], dim=0), dim=0)[0]
    return occs



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import einsum

from einops import rearrange

from utils.utils import bilinear_sampler, indexing


def nerf_encoding(x, L=6, NORMALIZE_FACOR=1 / 300):
    """
    x is of shape [*, 2]. The last dimension are two coordinates (x and y).
    """
    freq_bands = 2.0 ** torch.linspace(0, L, L - 1).to(x.device)
    return torch.cat(
        [
            x * NORMALIZE_FACOR,
            torch.sin(3.14 * x[..., -2:-1] * freq_bands * NORMALIZE_FACOR),
            torch.cos(3.14 * x[..., -2:-1] * freq_bands * NORMALIZE_FACOR),
            torch.sin(3.14 * x[..., -1:] * freq_bands * NORMALIZE_FACOR),
            torch.cos(3.14 * x[..., -1:] * freq_bands * NORMALIZE_FACOR),
        ],
        dim=-1,
    )


def sampler_gaussian(latent, mean, std, image_size, point_num=25):
    # latent [B, H*W, D]
    # mean [B, 2, H, W]
    # std [B, 1, H, W]
    H, W = image_size
    B, HW, D = latent.shape
    STD_MAX = 20
    latent = rearrange(
        latent, "b (h w) c -> b c h w", h=H, w=W
    )  # latent = latent.view(B, H, W, D).permute(0, 3, 1, 2)
    mean = mean.permute(0, 2, 3, 1)  # [B, H, W, 2]

    dx = torch.linspace(-1, 1, int(point_num**0.5))
    dy = torch.linspace(-1, 1, int(point_num**0.5))
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]
    delta_3sigma = (
        F.sigmoid(std.permute(0, 2, 3, 1).reshape(B * HW, 1, 1, 1))
        * STD_MAX
        * delta
        * 3
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    centroid = mean.reshape(B * H * W, 1, 1, 2)
    coords = centroid + delta_3sigma

    coords = rearrange(coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W)
    sampled_latents = bilinear_sampler(
        latent, coords
    )  # [B*H*W, dim, point_num**0.5, point_num**0.5]
    sampled_latents = sampled_latents.permute(0, 2, 3, 1)
    sampled_weights = -(torch.sum(delta.pow(2), dim=-1))

    return sampled_latents, sampled_weights


def sampler_gaussian_zy(
    latent, mean, std, image_size, point_num=25, return_deltaXY=False, beta=1
):
    # latent [B, H*W, D]
    # mean [B, 2, H, W]
    # std [B, 1, H, W]
    H, W = image_size
    B, HW, D = latent.shape
    latent = rearrange(
        latent, "b (h w) c -> b c h w", h=H, w=W
    )  # latent = latent.view(B, H, W, D).permute(0, 3, 1, 2)
    mean = mean.permute(0, 2, 3, 1)  # [B, H, W, 2]

    dx = torch.linspace(-1, 1, int(point_num**0.5))
    dy = torch.linspace(-1, 1, int(point_num**0.5))
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]
    delta_3sigma = (
        std.permute(0, 2, 3, 1).reshape(B * HW, 1, 1, 1) * delta * 3
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    centroid = mean.reshape(B * H * W, 1, 1, 2)
    coords = centroid + delta_3sigma

    coords = rearrange(coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W)
    sampled_latents = bilinear_sampler(
        latent, coords
    )  # [B*H*W, dim, point_num**0.5, point_num**0.5]
    sampled_latents = sampled_latents.permute(0, 2, 3, 1)
    sampled_weights = -(torch.sum(delta.pow(2), dim=-1)) / beta

    if return_deltaXY:
        return sampled_latents, sampled_weights, delta_3sigma
    else:
        return sampled_latents, sampled_weights


def sampler_gaussian(latent, mean, std, image_size, point_num=25, return_deltaXY=False):
    # latent [B, H*W, D]
    # mean [B, 2, H, W]
    # std [B, 1, H, W]
    H, W = image_size
    B, HW, D = latent.shape
    STD_MAX = 20
    latent = rearrange(
        latent, "b (h w) c -> b c h w", h=H, w=W
    )  # latent = latent.view(B, H, W, D).permute(0, 3, 1, 2)
    mean = mean.permute(0, 2, 3, 1)  # [B, H, W, 2]

    dx = torch.linspace(-1, 1, int(point_num**0.5))
    dy = torch.linspace(-1, 1, int(point_num**0.5))
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]
    delta_3sigma = (
        F.sigmoid(std.permute(0, 2, 3, 1).reshape(B * HW, 1, 1, 1))
        * STD_MAX
        * delta
        * 3
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    centroid = mean.reshape(B * H * W, 1, 1, 2)
    coords = centroid + delta_3sigma

    coords = rearrange(coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W)
    sampled_latents = bilinear_sampler(
        latent, coords
    )  # [B*H*W, dim, point_num**0.5, point_num**0.5]
    sampled_latents = sampled_latents.permute(0, 2, 3, 1)
    sampled_weights = -(torch.sum(delta.pow(2), dim=-1))

    if return_deltaXY:
        return sampled_latents, sampled_weights, delta_3sigma
    else:
        return sampled_latents, sampled_weights


def sampler_gaussian_fix(latent, mean, image_size, point_num=49):
    # latent [B, H*W, D]
    # mean [B, 2, H, W]
    H, W = image_size
    B, HW, D = latent.shape
    STD_MAX = 20
    latent = rearrange(
        latent, "b (h w) c -> b c h w", h=H, w=W
    )  # latent = latent.view(B, H, W, D).permute(0, 3, 1, 2)
    mean = mean.permute(0, 2, 3, 1)  # [B, H, W, 2]

    radius = int((int(point_num**0.5) - 1) / 2)

    dx = torch.linspace(-radius, radius, 2 * radius + 1)
    dy = torch.linspace(-radius, radius, 2 * radius + 1)
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    centroid = mean.reshape(B * H * W, 1, 1, 2)
    coords = centroid + delta

    coords = rearrange(coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W)
    sampled_latents = bilinear_sampler(
        latent, coords
    )  # [B*H*W, dim, point_num**0.5, point_num**0.5]
    sampled_latents = sampled_latents.permute(0, 2, 3, 1)
    sampled_weights = -(torch.sum(delta.pow(2), dim=-1)) / point_num  # smooth term

    return sampled_latents, sampled_weights


def sampler_gaussian_fix_pyramid(
    latent, feat_pyramid, scale_weight, mean, image_size, point_num=25
):
    # latent [B, H*W, D]
    # mean [B, 2, H, W]
    # scale weight [B, H*W, layer_num]

    H, W = image_size
    B, HW, D = latent.shape
    STD_MAX = 20
    latent = rearrange(
        latent, "b (h w) c -> b c h w", h=H, w=W
    )  # latent = latent.view(B, H, W, D).permute(0, 3, 1, 2)
    mean = mean.permute(0, 2, 3, 1)  # [B, H, W, 2]

    radius = int((int(point_num**0.5) - 1) / 2)

    dx = torch.linspace(-radius, radius, 2 * radius + 1)
    dy = torch.linspace(-radius, radius, 2 * radius + 1)
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    sampled_latents = []
    for i in range(len(feat_pyramid)):
        centroid = mean.reshape(B * H * W, 1, 1, 2)
        coords = (centroid + delta) / 2**i
        coords = rearrange(
            coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W
        )
        sampled_latents.append(bilinear_sampler(feat_pyramid[i], coords))

    sampled_latents = torch.stack(
        sampled_latents, dim=1
    )  # [B, layer_num, dim, H*W, point_num]
    sampled_latents = sampled_latents.permute(
        0, 3, 4, 2, 1
    )  # [B, H*W, point_num, dim, layer_num]
    scale_weight = F.softmax(scale_weight, dim=2)  # [B, H*W, layer_num]
    vis_out = scale_weight
    scale_weight = torch.unsqueeze(
        torch.unsqueeze(scale_weight, dim=2), dim=2
    )  # [B, HW, 1, 1, layer_num]

    weighted_latent = torch.sum(
        sampled_latents * scale_weight, dim=-1
    )  # [B, H*W, point_num, dim]

    sampled_weights = -(torch.sum(delta.pow(2), dim=-1)) / point_num  # smooth term

    return weighted_latent, sampled_weights, vis_out


def sampler_gaussian_pyramid(
    latent, feat_pyramid, scale_weight, mean, std, image_size, point_num=25
):
    # latent [B, H*W, D]
    # mean [B, 2, H, W]
    # scale weight [B, H*W, layer_num]

    H, W = image_size
    B, HW, D = latent.shape
    STD_MAX = 20
    latent = rearrange(
        latent, "b (h w) c -> b c h w", h=H, w=W
    )  # latent = latent.view(B, H, W, D).permute(0, 3, 1, 2)
    mean = mean.permute(0, 2, 3, 1)  # [B, H, W, 2]

    radius = int((int(point_num**0.5) - 1) / 2)

    dx = torch.linspace(-1, 1, int(point_num**0.5))
    dy = torch.linspace(-1, 1, int(point_num**0.5))
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]
    delta_3sigma = (
        std.permute(0, 2, 3, 1).reshape(B * HW, 1, 1, 1) * delta * 3
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    sampled_latents = []
    for i in range(len(feat_pyramid)):
        centroid = mean.reshape(B * H * W, 1, 1, 2)
        coords = (centroid + delta_3sigma) / 2**i
        coords = rearrange(
            coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W
        )
        sampled_latents.append(bilinear_sampler(feat_pyramid[i], coords))

    sampled_latents = torch.stack(
        sampled_latents, dim=1
    )  # [B, layer_num, dim, H*W, point_num]
    sampled_latents = sampled_latents.permute(
        0, 3, 4, 2, 1
    )  # [B, H*W, point_num, dim, layer_num]
    scale_weight = F.softmax(scale_weight, dim=2)  # [B, H*W, layer_num]
    vis_out = scale_weight
    scale_weight = torch.unsqueeze(
        torch.unsqueeze(scale_weight, dim=2), dim=2
    )  # [B, HW, 1, 1, layer_num]

    weighted_latent = torch.sum(
        sampled_latents * scale_weight, dim=-1
    )  # [B, H*W, point_num, dim]

    sampled_weights = -(torch.sum(delta.pow(2), dim=-1)) / point_num  # smooth term

    return weighted_latent, sampled_weights, vis_out


def sampler_gaussian_fix_MH(latent, mean, image_size, point_num=25):
    """different heads have different mean"""
    # latent [B, H*W, D]
    # mean [B, 2, H, W, heands]

    H, W = image_size
    B, HW, D = latent.shape
    _, _, _, _, HEADS = mean.shape
    STD_MAX = 20
    latent = rearrange(latent, "b (h w) c -> b c h w", h=H, w=W)
    mean = mean.permute(0, 2, 3, 4, 1)  # [B, H, W, heads, 2]

    radius = int((int(point_num**0.5) - 1) / 2)

    dx = torch.linspace(-radius, radius, 2 * radius + 1)
    dy = torch.linspace(-radius, radius, 2 * radius + 1)
    delta = (
        torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1)
        .to(mean.device)
        .repeat(HEADS, 1, 1, 1)
    )  # [HEADS, point_num**0.5, point_num**0.5, 2]

    centroid = mean.reshape(B * H * W, HEADS, 1, 1, 2)
    coords = centroid + delta
    coords = rearrange(
        coords, "(b h w) H r1 r2 c -> b (h w H) (r1 r2) c", b=B, h=H, w=W, H=HEADS
    )
    sampled_latents = bilinear_sampler(latent, coords)  # [B, dim, H*W*HEADS, pointnum]
    sampled_latents = sampled_latents.permute(
        0, 2, 3, 1
    )  # [B, H*W*HEADS, pointnum, dim]
    sampled_weights = -(torch.sum(delta.pow(2), dim=-1)) / point_num  # smooth term
    return sampled_latents, sampled_weights


def sampler_gaussian_fix_pyramid_MH(
    latent, feat_pyramid, scale_head_weight, mean, image_size, point_num=25
):
    # latent [B, H*W, D]
    # mean [B, 2, H, W, heands]
    # scale_head weight [B, H*W, layer_num*heads]

    H, W = image_size
    B, HW, D = latent.shape
    _, _, _, _, HEADS = mean.shape

    latent = rearrange(latent, "b (h w) c -> b c h w", h=H, w=W)
    mean = mean.permute(0, 2, 3, 4, 1)  # [B, H, W, heads, 2]

    radius = int((int(point_num**0.5) - 1) / 2)

    dx = torch.linspace(-radius, radius, 2 * radius + 1)
    dy = torch.linspace(-radius, radius, 2 * radius + 1)
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        mean.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    sampled_latents = []
    centroid = mean.reshape(B * H * W, HEADS, 1, 1, 2)
    for i in range(len(feat_pyramid)):
        coords = (centroid) / 2**i + delta
        coords = rearrange(
            coords, "(b h w) H r1 r2 c -> b (h w H) (r1 r2) c", b=B, h=H, w=W, H=HEADS
        )
        sampled_latents.append(
            bilinear_sampler(feat_pyramid[i], coords)
        )  # [B, dim, H*W*HEADS, point_num]

    sampled_latents = torch.stack(
        sampled_latents, dim=1
    )  # [B, layer_num, dim, H*W*HEADS, point_num]
    sampled_latents = sampled_latents.permute(
        0, 3, 4, 2, 1
    )  # [B, H*W*HEADS, point_num, dim, layer_num]

    scale_head_weight = scale_head_weight.reshape(B, H * W * HEADS, -1)
    scale_head_weight = F.softmax(scale_head_weight, dim=2)  # [B, H*W*HEADS, layer_num]
    scale_head_weight = torch.unsqueeze(
        torch.unsqueeze(scale_head_weight, dim=2), dim=2
    )  # [B, H*W*HEADS, 1, 1, layer_num]

    weighted_latent = torch.sum(
        sampled_latents * scale_head_weight, dim=-1
    )  # [B, H*W*HEADS, point_num, dim]

    sampled_weights = -(torch.sum(delta.pow(2), dim=-1)) / point_num  # smooth term

    return weighted_latent, sampled_weights


def sampler(feat, center, window_size):
    # feat [B, C, H, W]
    # center [B, 2, H, W]
    center = center.permute(0, 2, 3, 1)  # [B, H, W, 2]
    B, H, W, C = center.shape

    radius = window_size // 2
    dx = torch.linspace(-radius, radius, 2 * radius + 1)
    dy = torch.linspace(-radius, radius, 2 * radius + 1)
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        center.device
    )  # [B*H*W, window_size, point_num**0.5, 2]

    center = center.reshape(B * H * W, 1, 1, 2)
    coords = center + delta

    coords = rearrange(coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W)
    sampled_latents = bilinear_sampler(
        feat, coords
    )  # [B*H*W, dim, window_size, window_size]
    # sampled_latents = sampled_latents.permute(0, 2, 3, 1)

    return sampled_latents


def retrieve_tokens(feat, center, window_size, sampler):
    # feat [B, C, H, W]
    # center [B, 2, H, W]
    radius = window_size // 2
    dx = torch.linspace(-radius, radius, 2 * radius + 1)
    dy = torch.linspace(-radius, radius, 2 * radius + 1)
    delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
        center.device
    )  # [B*H*W, point_num**0.5, point_num**0.5, 2]

    B, H, W, C = center.shape
    centroid = center.reshape(B * H * W, 1, 1, 2)
    coords = centroid + delta

    coords = rearrange(coords, "(b h w) r1 r2 c -> b (h w) (r1 r2) c", b=B, h=H, w=W)
    if sampler == "nn":
        sampled_latents = indexing(feat, coords)
    elif sampler == "bilinear":
        sampled_latents = bilinear_sampler(feat, coords)
    else:
        raise ValueError("invalid sampler")
    # [B, dim, H*W, point_num]

    return sampled_latents


def pyramid_retrieve_tokens(
    feat_pyramid, center, image_size, window_sizes, sampler="bilinear"
):
    center = center.permute(0, 2, 3, 1)  # [B, H, W, 2]
    sampled_latents_pyramid = []
    for idx in range(len(window_sizes)):
        sampled_latents_pyramid.append(
            retrieve_tokens(feat_pyramid[idx], center, window_sizes[idx], sampler)
        )
        center = center / 2

    return torch.cat(sampled_latents_pyramid, dim=-1)


class FeedForward(nn.Module):
    def __init__(self, dim, dropout=0.0):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(dim, dim),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        x = self.net(x)
        return x


class MLP(nn.Module):
    def __init__(self, in_dim=22, out_dim=1, innter_dim=96, depth=5):
        super().__init__()
        self.FC1 = nn.Linear(in_dim, innter_dim)
        self.FC_out = nn.Linear(innter_dim, out_dim)
        self.relu = torch.nn.LeakyReLU(0.2)
        self.FC_inter = nn.ModuleList(
            [nn.Linear(innter_dim, innter_dim) for i in range(depth)]
        )

    def forward(self, x):
        x = self.FC1(x)
        x = self.relu(x)
        for inter_fc in self.FC_inter:
            x = inter_fc(x)
            x = self.relu(x)
        x = self.FC_out(x)
        return x


class MultiHeadAttention(nn.Module):
    def __init__(self, dim, heads, num_kv_tokens, cfg, rpe_bias=None, use_rpe=False):
        super(MultiHeadAttention, self).__init__()
        self.dim = dim
        self.heads = heads
        self.num_kv_tokens = num_kv_tokens
        self.scale = (dim / heads) ** -0.5
        self.rpe = cfg.rpe
        self.attend = nn.Softmax(dim=-1)
        self.use_rpe = use_rpe

        if use_rpe:
            if rpe_bias is None:
                if self.rpe == "element-wise":
                    self.rpe_bias = nn.Parameter(
                        torch.zeros(heads, self.num_kv_tokens, dim // heads)
                    )
                elif self.rpe == "head-wise":
                    self.rpe_bias = nn.Parameter(
                        torch.zeros(1, heads, 1, self.num_kv_tokens)
                    )
                elif self.rpe == "token-wise":
                    self.rpe_bias = nn.Parameter(
                        torch.zeros(1, 1, 1, self.num_kv_tokens)
                    )  # 81 is point_num
                elif self.rpe == "implicit":
                    pass
                    # self.implicit_pe_fn = MLP(in_dim=22, out_dim=self.dim, innter_dim=int(self.dim//2.4), depth=2)
                    # raise ValueError('Implicit Encoding Not Implemented')
                elif self.rpe == "element-wise-value":
                    self.rpe_bias = nn.Parameter(
                        torch.zeros(heads, self.num_kv_tokens, dim // heads)
                    )
                    self.rpe_value = nn.Parameter(torch.randn(self.num_kv_tokens, dim))
                else:
                    raise ValueError("Not Implemented")
            else:
                self.rpe_bias = rpe_bias

    def attend_with_rpe(self, Q, K, rpe_bias):
        Q = rearrange(Q, "b i (heads d) -> b heads i d", heads=self.heads)
        K = rearrange(K, "b j (heads d) -> b heads j d", heads=self.heads)

        dots = (
            einsum("bhid, bhjd -> bhij", Q, K) * self.scale
        )  # (b hw) heads 1 pointnum
        if self.use_rpe:
            if self.rpe == "element-wise":
                rpe_bias_weight = (
                    einsum("bhid, hjd -> bhij", Q, rpe_bias) * self.scale
                )  # (b hw) heads 1 pointnum
                dots = dots + rpe_bias_weight
            elif self.rpe == "implicit":
                pass
                rpe_bias_weight = (
                    einsum("bhid, bhjd -> bhij", Q, rpe_bias) * self.scale
                )  # (b hw) heads 1 pointnum
                dots = dots + rpe_bias_weight
            elif self.rpe == "head-wise" or self.rpe == "token-wise":
                dots = dots + rpe_bias

        return self.attend(dots), dots

    def forward(self, Q, K, V, rpe_bias=None):
        if self.use_rpe:
            if rpe_bias is None or self.rpe == "element-wise":
                rpe_bias = self.rpe_bias
            else:
                rpe_bias = rearrange(
                    rpe_bias, "b hw pn (heads d) -> (b hw) heads pn d", heads=self.heads
                )
            attn, dots = self.attend_with_rpe(Q, K, rpe_bias)
        else:
            attn, dots = self.attend_with_rpe(Q, K, None)
        B, HW, _ = Q.shape

        if V is not None:
            V = rearrange(V, "b j (heads d) -> b heads j d", heads=self.heads)

            out = einsum("bhij, bhjd -> bhid", attn, V)
            out = rearrange(out, "b heads hw d -> b hw (heads d)", b=B, hw=HW)
        else:
            out = None

        # dots = torch.squeeze(dots, 2)
        # dots = rearrange(dots, '(b hw) heads d -> b hw (heads d)', b=B, hw=HW)

        return out, dots
